@@ -42,9 +42,7 @@ Database::Database()
 
     // speed up database (especially for slow persistent memory like on the PinePhone)
     execute(QStringLiteral("PRAGMA synchronous = OFF;"));
-    execute(QStringLiteral("PRAGMA journal_mode = WAL;")); // TODO: or MEMORY?
-    execute(QStringLiteral("PRAGMA temp_store = MEMORY;"));
-    execute(QStringLiteral("PRAGMA locking_mode = EXCLUSIVE;"));
+    execute(QStringLiteral("PRAGMA journal_mode = WAL;")); // no measurable effect in Database benchmark, but recommended everywhere
 
     // prepare queries once (faster)
     m_addGroupQuery.reset(new QSqlQuery(db));
@@ -101,7 +99,7 @@ Database::Database()
     m_addProgramCategoryQuery.reset(new QSqlQuery(db));
     success &= m_addProgramCategoryQuery->prepare(QStringLiteral("INSERT OR IGNORE INTO ProgramCategories VALUES (:program, :category);"));
     m_programCategoriesQuery.reset(new QSqlQuery(db));
-    success &= m_programCategoriesQuery->prepare(QStringLiteral("SELECT category FROM ProgramCategories WHERE program=:program;"));
+    success &= m_programCategoriesQuery->prepare(QStringLiteral("SELECT * FROM ProgramCategories;"));
 
     if (!success) {
         qCritical() << "Failed to prepare database queries";
@@ -232,14 +230,24 @@ void Database::cleanup()
     QDateTime dateTimeFuture = QDateTime::currentDateTime();
     dateTimeFuture = dateTimeFuture.addDays(30);
 
+    QSqlDatabase::database().transaction();
     QSqlQuery query;
     if (!query.prepare(QStringLiteral("DELETE FROM Programs WHERE stop < :sinceEpochPast OR stop > :sinceEpochFuture;"))) {
-        qCritical() << "Failed to prepare cleanup query";
+        qCritical() << "Failed to prepare cleanup query for Programs";
         return;
     }
     query.bindValue(QStringLiteral(":sinceEpochPast"), dateTimePast.toSecsSinceEpoch());
     query.bindValue(QStringLiteral(":sinceEpochFuture"), dateTimeFuture.toSecsSinceEpoch());
     execute(query);
+
+    // delete categories which are no longer referenced
+    if (!query.prepare(QStringLiteral("DELETE FROM ProgramCategories WHERE program NOT IN (SELECT id FROM Programs WHERE id IS NOT NULL);"))) {
+        qCritical() << "Failed to prepare cleanup query for ProgramCategories";
+        return;
+    }
+    execute(query);
+
+    QSqlDatabase::database().commit();
 }
 
 void Database::addGroup(const GroupData &data)
@@ -331,10 +339,9 @@ void Database::addChannel(const ChannelData &data, const GroupId &group)
 
         // store channel
         {
-            QUrl urlFromInput = QUrl::fromUserInput(data.m_url);
             m_addChannelQuery->bindValue(QStringLiteral(":id"), data.m_id.value());
             m_addChannelQuery->bindValue(QStringLiteral(":name"), data.m_name);
-            m_addChannelQuery->bindValue(QStringLiteral(":url"), urlFromInput.toString());
+            m_addChannelQuery->bindValue(QStringLiteral(":url"), data.m_url);
             m_addChannelQuery->bindValue(QStringLiteral(":group"), group.value());
             m_addChannelQuery->bindValue(QStringLiteral(":image"), data.m_image);
             execute(*m_addChannelQuery);
@@ -565,6 +572,9 @@ QMap<ChannelId, QVector<ProgramData>> Database::programs() const
 {
     QMap<ChannelId, QVector<ProgramData>> programs;
 
+    QSqlDatabase::database().transaction();
+    const QMultiMap<ProgramId, QString> categories = programCategories();
+
     execute(*m_programsQuery);
 
     while (m_programsQuery->next()) {
@@ -584,15 +594,13 @@ QMap<ChannelId, QVector<ProgramData>> Database::programs() const
         data.m_description = m_programsQuery->value(QStringLiteral("description")).toString();
         data.m_descriptionFetched = m_programsQuery->value(QStringLiteral("descriptionFetched")).toBool();
 
-        m_programCategoriesQuery->bindValue(QStringLiteral(":program"), data.m_id.value());
-        execute(*m_programCategoriesQuery);
-
-        while (m_programCategoriesQuery->next()) {
-            data.m_categories.push_back(m_programCategoriesQuery->value(QStringLiteral("category")).toString());
+        for (auto &&category : categories.values(data.m_id)) {
+            data.m_categories.push_back(category);
         }
 
         programs[channelId].push_back(data);
     }
+    QSqlDatabase::database().commit();
 
     return programs;
 }
@@ -601,6 +609,9 @@ QVector<ProgramData> Database::programs(const ChannelId &channelId) const
 {
     QVector<ProgramData> programs;
 
+    QSqlDatabase::database().transaction();
+    const QMultiMap<ProgramId, QString> categories = programCategories();
+
     m_programsPerChannelQuery->bindValue(QStringLiteral(":channel"), channelId.value());
     execute(*m_programsPerChannelQuery);
 
@@ -616,14 +627,24 @@ QVector<ProgramData> Database::programs(const ChannelId &channelId) const
         data.m_description = m_programsPerChannelQuery->value(QStringLiteral("description")).toString();
         data.m_descriptionFetched = m_programsPerChannelQuery->value(QStringLiteral("descriptionFetched")).toBool();
 
-        m_programCategoriesQuery->bindValue(QStringLiteral(":program"), data.m_id.value());
-        execute(*m_programCategoriesQuery);
-
-        while (m_programCategoriesQuery->next()) {
-            data.m_categories.push_back(m_programCategoriesQuery->value(QStringLiteral("category")).toString());
+        for (auto &&category : categories.values(data.m_id)) {
+            data.m_categories.push_back(category);
         }
 
         programs.push_back(data);
     }
+    QSqlDatabase::database().commit();
+
     return programs;
 }
+
+QMultiMap<ProgramId, QString> Database::programCategories() const
+{
+    QMultiMap<ProgramId, QString> categories;
+    execute(*m_programCategoriesQuery);
+    while (m_programCategoriesQuery->next()) {
+        categories.insert(ProgramId(m_programCategoriesQuery->value(QStringLiteral("program")).toString()),
+                          m_programCategoriesQuery->value(QStringLiteral("category")).toString());
+    }
+    return categories;
+}
